using LinearAlgebra
using Arpack
using SparseArrays
using KrylovKit

eye(n) = Matrix{Float64}(I,n,n)
function eig(H) 
    res = eigen(H)
    res.values,res.vectors
end
# The lines above revive some Julia 0.6 functions when using Julia 1.0

#Julia eigenvalue routine autodetects symmetric matrices, but only if exactly symmetric
function symmetrize!(m::Array{Float64,2})
    n = size(m,1)
    for i=1:n, j=i:n
	m[i,j] = m[j,i] = 0.5 * (m[i,j]+m[j,i])
    end
end
function eigsym(H)
    symmetrize!(H)
    eig(H)
end

sz = Float64[0.5 0; 
	     0 -0.5]
sp = Float64[0 1; 
	     0 0]
sm = sp'

H2 = [0.5 * (sp[s1p,s1] * sm[s2p,s2] + sm[s1p,s1] * sp[s2p,s2]) + sz[s1p,s1] * sz[s2p,s2] 
	    for s1p=1:2,s2p=1:2, s1=1:2,s2=1:2]
Hmat2 = reshape(H2,4,4)
symmetrize!(Hmat2)
evals,evecs = eigsym(Hmat2)
println("H2 eigenvalues: ",evals)

one = eye(2)	# 2x2 identity matrix
H4 = [
     H2[s1p,s2p,s1,s2] * one[s3p,s3] * one[s4p,s4] + 
     one[s1p,s1] * H2[s2p,s3p,s2,s3] * one[s4p,s4] + 
     one[s1p,s1] * one[s2p,s2] * H2[s3p,s4p,s3,s4]
    for s1p=1:2, s2p=1:2, s3p=1:2, s4p=1:2,
	 s1=1:2,  s2=1:2,  s3=1:2,  s4=1:2]

H4sp = sparse(reshape(H4,16,16))

#eigs uses Lanczos, in Arpack
res = eigs(H4sp,nev=1, which=:SR,ritzvec=true,maxiter=10000)
@show res[1]
@show res[2]

# From KrylovKit
resk = eigsolve(H4sp,1,:SR)
energy = resk[1][1]
@show energy
ground = resk[2][1]
@show ground
@show dot(res[2],ground)
