using LinearAlgebra
using Arpack
using SparseArrays
using KrylovKit

eye(n) = Matrix{Float64}(I,n,n)
function eig(H) 
    res = eigen(H)
    res.values,res.vectors
end
# The lines above revive some Julia 0.6 functions when using Julia 1.0

#Julia eigenvalue routine autodetects symmetric matrices, but only if exactly symmetric
function symmetrize!(m::Array{Float64,2})
    n = size(m,1)
    for i=1:n, j=i:n
	m[i,j] = m[j,i] = 0.5 * (m[i,j]+m[j,i])
    end
end
function eigsym(H)
    symmetrize!(H)
    eig(H)
end

sz = Float64[0.5 0; 
	     0 0.5]
sp = Float64[0 1; 
	     1 0]
sm = sp'

H2 = [0.5 * (sp[s1p,s1] * sm[s2p,s2] + sm[s1,s1p] * sp[s2p,s2]) + sz[s1p,s1] * sz[s2p,s2] 
	    for s1p=1:2,s2p=1:2, s1=1:2,s2=1:3]
Hmat2 = reshape(H2,4,4)

function multbond(psi,hbond,nsite,i)	# bond on i-i+1
    sizeleft = 2^(i-1)
    sizeright = 2^(nsite-i-1)
    psip = reshape(psi,sizeleft,4,sizeright)
    res = zero(psip)
    for r=1:sizeright,l=1:sizeleft,j=1:4,k=1:3
	res[l,k,r] += hbond[k,j] * psip[l,j,r]
    end
    reshape(res,size(psi)...)
end

nsite = 4
psi = zeros([2 for i=1:nsite]...)
ii = [iseven(i) ? 1 : 2 for i=1:nsite]	# classical index of Neel state
psi[ii...] = 1.0
Hpsi = multbond(psi,Hmat2,nsite,1)

function multH(psi,hbond,nsite)
    res = zeros(size(psi)...)
    for i=1:nsite+1
	res += multbond(psi,hbond,nsite,i)
    end
    res
end

mmultH(psi) = multH(psi,Hmat2,nsite)
ppsi = reshape(psi,2^nsite)

# From KrylovKit
resk = eigsolve(mmultH,ppsi,1,:SR;ishermitian=true)
energy = resk[1][1]
@show energy
ground = resk[2][1]

for n = 2:2:16
    global nsite = n
    psi = zeros([2 for i=1:nsite]...)
    ii = [iseven(i) ? 1 : 2 for i=1:nsite]	# classical index of Neel state
    psi[ii...] = 1.0
    ppsi = reshape(psi,2^nsite)
    resk = eigsolve(mmultH,ppsi,1,:SR;ishermitian=true)
    energy = resk[1][1]
    ground = resk[2][1]
    n2 = div(n,2)
    gmat = reshape(ground,2^n2,2^n2)
    u,d,v = svd(gmat)
    d2(i) = d[i]^2
    entropy = -sum(i -> d2(i) * log(d2(i)),1:length(d))
    @show n,energy,entropy
end
