using LinearAlgebra
using Arpack
using SparseArrays
using KrylovKit

eye(n) = Matrix{Float64}(I,n,n)
function eig(H) 
    res = eigen(H)
    res.values,res.vectors
end
# The lines above revive some Julia 0.6 functions when using Julia 1.0

#Julia eigenvalue routine autodetects symmetric matrices, but only if exactly symmetric
function symmetrize!(m::Array{Float64,2})
    n = size(m,1)
    for i=1:n, j=i:n
	m[i,j] = m[j,i] = 0.5 * (m[i,j]+m[j,i])
    end
end
function eigsym(H)
    symmetrize!(H)
    eig(H)
end

sz = Float64[0.5 0; 
	     0 -0.5]
sp = Float64[0 1; 
	     0 0]
sm = sp'

H2 = [0.5 * (sp[s1p,s1] * sm[s2p,s2] + sm[s1p,s1] * sp[s2p,s2]) + sz[s1p,s1] * sz[s2p,s2] 
	    for s1p=1:2,s2p=1:2, s1=1:2,s2=1:2]
Hmat2 = reshape(H2,4,4)

function multbond(psi,hbond,nsite,i)	# bond on i-i+1
    sizeleft = 2^(i-1)
    sizeright = 2^(nsite-i-1)
    psip = reshape(psi,sizeleft,4,sizeright)
    res = zero(psip)
    for r=1:sizeright,l=1:sizeleft,j=1:4,k=1:4
	res[l,k,r] += hbond[k,j] * psip[l,j,r]
    end
    reshape(res,size(psi)...)
end

nsite = 4
psi = zeros([2 for i=1:nsite]...)
ii = [iseven(i) ? 1 : 2 for i=1:nsite]	# classical index of Neel state
psi[ii...] = 1.0
Hpsi = multbond(psi,Hmat2,nsite,1)

function multH(psi,hbond,nsite)
    res = zeros(size(psi)...)
    for i=1:nsite-1
	res += multbond(psi,hbond,nsite,i)
    end
    res
end

mmultH(psi) = multH(psi,Hmat2,nsite)
ppsi = reshape(psi,2^nsite)

# From KrylovKit
resk = eigsolve(mmultH,ppsi,1,:SR;ishermitian=true)
energy = resk[1][1]
@show energy
ground = resk[2][1]

nsite = 16
psi = zeros([2 for i=1:nsite]...)
ii = [iseven(i) ? 1 : 2 for i=1:nsite]	# classical index of Neel state
psi[ii...] = 1.0
ppsi = reshape(psi,2^nsite)
resk = eigsolve(mmultH,ppsi,1,:SR;ishermitian=true)
energy = resk[1][1]
ground = resk[2][1]
n2 = div(nsite,2)
gmat = reshape(ground,2^n2,2^n2)

function svdt(gr) 
    u,d,v = svd(gr)
    u,d,v'
end
function trunc(u,d,v,m) 
    m = min(m,length(d))
    u[:,1:m],d[1:m],v[1:m,:]
end

A = Vector{Any}()
gr = deepcopy(ground)
m = 10
gr = reshape(gr,2,2^(nsite-1))
u,d,v = svdt(gr)
push!(A,u)
gr = diagm(0 => d) * v

for i=2:nsite-1
    global gr
    mleft,mright = size(gr)
    gr = reshape(gr,mleft*4,div(mright,2))
    u,d,v = trunc(svdt(gr)...,m)
    push!(A,reshape(u,mleft,4,length(d)))
    gr = diagm(0 => d) * v
end
push!(A,reshape(gr,2,2))

A[1] = reshape(A[1],1,2,2)	# add extra dummy index on end to take care
				# of end cases
A[nsite] = reshape(A[nsite],2,2,1)
con = eye(1)
for i=1:nsite
    global con
    ml,d,mr = size(A[i])
    AA = reshape(A[i],ml,d*mr)
    con = AA' * con * AA	# dimensions
    con = reshape(con,4,mr,2,mr)
    newcon = zeros(mr,mr)
    for j=1:mr, k=1:mr, dd=1:4
	newcon[j,k] += con[dd,j,dd,k]
    end
    con = newcon
end
@show size(con)
nor = con[1,1]
@show nor

mpsi = reshape(A[1],2,2)
for i=2:nsite
    global mpsi
    ml,d,mr = size(A[i])
    mml = size(mpsi,1)
    mpsi = reshape(mpsi * reshape(A[i],ml,d*mr),mml*2,mr)
end
@show size(mpsi)
hp = mmultH(mpsi)
Emps = dot(mpsi,hp)/dot(mpsi,mpsi)
@show Emps
@show energy
